@@ -250,14 +250,13 @@ where
         phy_addr: Option<u32>,
         clk_config: &RmiiClockConfig,
     ) -> Result<(*mut esp_eth_mac_t, *mut esp_eth_phy_t), EspError> {
-        let mac_cfg = EspEth::<RmiiEthPeripherals<MDC, MDIO>>::eth_mac_rmii_config(clk_config);
+        let mac = EspEth::<RmiiEthPeripherals<MDC, MDIO>>::eth_mac_new(clk_config);
+
         let phy_cfg = EspEth::<RmiiEthPeripherals<MDC, MDIO>>::eth_phy_default_config(
             reset.as_ref().map(|p| p.pin()),
             phy_addr,
         );
 
-        let mac = unsafe { esp_eth_mac_new_esp32(&mac_cfg) };
-
         let phy = match chipset {
             RmiiEthChipset::IP101 => unsafe { esp_eth_phy_new_ip101(&phy_cfg) },
             RmiiEthChipset::RTL8201 => unsafe { esp_eth_phy_new_rtl8201(&phy_cfg) },
@@ -277,7 +276,8 @@ where
         Ok((mac, phy))
     }
 
-    fn eth_mac_rmii_config(clk_config: &RmiiClockConfig) -> eth_mac_config_t {
+    #[cfg(esp_idf_version_major = "4")]
+    fn eth_mac_new(clk_config: &RmiiClockConfig) -> *mut esp_eth_mac_t {
         let mut config = Self::eth_mac_default_config();
 
         #[cfg(not(esp_idf_version = "4.3"))]
@@ -285,7 +285,17 @@ where
             config.clock_config = clk_config.eth_mac_clock_config();
         }
 
-        config
+        unsafe { esp_eth_mac_new_esp32(&config) }
+    }
+
+    #[cfg(not(esp_idf_version_major = "4"))]
+    fn eth_mac_new(clk_config: &RmiiClockConfig) -> *mut esp_eth_mac_t {
+        let mut esp32_config = Self::eth_esp32_emac_default_config();
+        esp32_config.clock_config = clk_config.eth_mac_clock_config();
+
+        let config = Self::eth_mac_default_config();
+
+        unsafe { esp_eth_mac_new_esp32(&esp32_config, &config) }
     }
 }
 
@@ -885,6 +895,7 @@ impl<P> EspEth<P> {
         }
     }
 
+    #[cfg(esp_idf_version_major = "4")]
     fn eth_mac_default_config() -> eth_mac_config_t {
         eth_mac_config_t {
             sw_reset_timeout_ms: 100,
@@ -893,7 +904,28 @@ impl<P> EspEth<P> {
             smi_mdc_gpio_num: 23,
             smi_mdio_gpio_num: 18,
             flags: 0,
-            #[cfg(any(esp_idf_version = "4.4", esp_idf_version_major = "5"))]
+            #[cfg(esp_idf_version = "4.4")]
+            interface: eth_data_interface_t_EMAC_DATA_INTERFACE_RMII,
+            ..Default::default()
+        }
+    }
+
+    #[cfg(not(esp_idf_version_major = "4"))]
+    fn eth_mac_default_config() -> eth_mac_config_t {
+        eth_mac_config_t {
+            sw_reset_timeout_ms: 100,
+            rx_task_stack_size: 2048,
+            rx_task_prio: 15,
+            flags: 0,
+            ..Default::default()
+        }
+    }
+
+    #[cfg(not(esp_idf_version_major = "4"))]
+    fn eth_esp32_emac_default_config() -> eth_esp32_emac_config_t {
+        eth_esp32_emac_config_t {
+            smi_mdc_gpio_num: 23,
+            smi_mdio_gpio_num: 18,
             interface: eth_data_interface_t_EMAC_DATA_INTERFACE_RMII,
             ..Default::default()
         }
