@@ -265,27 +265,32 @@ impl EspWifi {
         sys_loop_stack: Arc<EspSysLoopStack>,
         nvs: Arc<EspDefaultNvs>,
     ) -> Result<Self, EspError> {
+        #[allow(clippy::needless_update)]
         let cfg = wifi_init_config_t {
+            #[cfg(esp_idf_version_major = "4")]
             event_handler: Some(esp_event_send_internal),
             osi_funcs: unsafe { &mut g_wifi_osi_funcs },
             wpa_crypto_funcs: unsafe { g_wifi_default_wpa_crypto_funcs },
-            static_rx_buf_num: 10,
-            dynamic_rx_buf_num: 32,
-            tx_buf_type: 1,
-            static_tx_buf_num: 0,
-            dynamic_tx_buf_num: 32,
-            csi_enable: 0,
-            ampdu_rx_enable: 1,
-            ampdu_tx_enable: 1,
-            nvs_enable: 0,
-            nano_enable: 0,
-            //tx_ba_win: 6,
-            rx_ba_win: 6,
-            wifi_task_core_id: 0,
-            beacon_max_len: 752,
-            mgmt_sbuf_num: 32,
-            feature_caps: 1, // CONFIG_FEATURE_WPA3_SAE_BIT
-            magic: 0x1F2F3F4F,
+            static_rx_buf_num: CONFIG_ESP32_WIFI_STATIC_RX_BUFFER_NUM as _,
+            dynamic_rx_buf_num: CONFIG_ESP32_WIFI_DYNAMIC_RX_BUFFER_NUM as _,
+            tx_buf_type: CONFIG_ESP32_WIFI_TX_BUFFER_TYPE as _,
+            static_tx_buf_num: WIFI_STATIC_TX_BUFFER_NUM as _,
+            dynamic_tx_buf_num: WIFI_DYNAMIC_TX_BUFFER_NUM as _,
+            cache_tx_buf_num: WIFI_CACHE_TX_BUFFER_NUM as _,
+            csi_enable: WIFI_CSI_ENABLED as _,
+            ampdu_rx_enable: WIFI_AMPDU_RX_ENABLED as _,
+            ampdu_tx_enable: WIFI_AMPDU_TX_ENABLED as _,
+            amsdu_tx_enable: WIFI_AMSDU_TX_ENABLED as _,
+            nvs_enable: WIFI_NVS_ENABLED as _,
+            nano_enable: WIFI_NANO_FORMAT_ENABLED as _,
+            //tx_ba_win: WIFI_DEFAULT_TX_BA_WIN as _,
+            rx_ba_win: WIFI_DEFAULT_RX_BA_WIN as _,
+            wifi_task_core_id: WIFI_TASK_CORE_ID as _,
+            beacon_max_len: WIFI_SOFTAP_BEACON_MAX_LEN as _,
+            mgmt_sbuf_num: WIFI_MGMT_SBUF_NUM as _,
+            feature_caps: unsafe { g_wifi_feature_caps },
+            sta_disconnected_pm: WIFI_STA_DISCONNECTED_PM_ENABLED != 0,
+            magic: WIFI_INIT_CONFIG_MAGIC as _,
             ..Default::default()
         };
         esp!(unsafe { esp_wifi_init(&cfg) })?;
